from collections.abc import Sequence
from typing import Generic, TypeVar

import attrs
import networkx as nx
import numpy as np
from numpy.ma import MaskedArray

TNumber = TypeVar("TNumber", bound=np.number)


@attrs.frozen(kw_only=True)
class Cycles(Generic[TNumber]):
    """Cycles of the eigenvalues."""

    cycles: list[np.ndarray[tuple[int, int], np.dtype[TNumber]]]
    """Cycles of the eigenvalues of length num_cycles.

    Ordered by length in descending order of shape (cycle_length, num_points)."""
    graph: nx.DiGraph
    """Graph of the cycles. If eigvals[i, -1] ~ eigvals[j, 0], then
    there is an edge from i to j."""
    eigvals: np.ndarray[tuple[int, int], np.dtype[TNumber]]
    """Continuous eigenvalues of shape (num_complete_eigenvalues, num_points)."""
    incomplete_eigvals: MaskedArray[tuple[int, int], np.dtype[TNumber]]
    """Continuous eigenvalues of shape (num_eigenvalues, num_points)
    with some NaN values."""

    @property
    def connected_cycles(self) -> list[np.ndarray[tuple[int], np.dtype[TNumber]]]:
        """
        List of connected sequences of eigenvalues of length num_cycles.

        Prdered by length in descending order same as self.cycles
        of shape (cycle_length, num_points).
        """
        return [
            np.stack(
                [np.concatenate(np.roll(cycle, -i, axis=0)) for i in range(cycle.shape[0])],
                axis=0,
            )
            for cycle in self.cycles
        ]

    @property
    def cycle_lengths(self) -> list[int]:
        """Lengths of the cycles in descending order, same as self.cycles."""
        return [cycle.shape[0] for cycle in self.cycles]

    @property
    def max_cycle_length(self) -> int:
        """Maximum length of the cycles."""
        if len(self.cycle_lengths) == 0:
            return 0
        return np.max(self.cycle_lengths)


def argmatch_from_closest(
    x: np.ndarray[tuple[int], np.dtype[TNumber]],
    y: np.ndarray[tuple[int], np.dtype[TNumber]],
    /,
) -> MaskedArray[tuple[int], np.dtype[TNumber]]:
    """
    Match the elements of x and y from the closest without duplicates.

    Parameters
    ----------
    x : np.ndarray[tuple[int], np.dtype[TNumber]]
        First array of shape (n,).
    y : np.ndarray[tuple[int], np.dtype[TNumber]]
        Second array of shape (m,).

    Returns
    -------
    MaskedArray[tuple[int], np.dtype[TNumber]]
        Array of shape (n,).
        (x[i], y[result[i]]) is the pair.

    """
    if x.ndim != 1:
        raise ValueError(f"{x.ndim=} should be 1")
    if y.ndim != 1:
        raise ValueError(f"{y.ndim=} should be 1")
    dist = np.ma.masked_array(np.abs(x[:, None] - y[None, :]), mask=False)
    result = np.full_like(x, -1, dtype=int)
    for _ in range(min(len(x), len(y))):
        idx = np.unravel_index(dist.argmin(), dist.shape)
        result[idx[0]] = idx[1]
        dist.mask[idx[0], :] = np.inf
        dist.mask[:, idx[1]] = np.inf
    result = np.ma.masked_array(result, mask=result == -1)
    return result


def argmatch_from_closest_masked(
    x: MaskedArray[tuple[int], np.dtype[TNumber]],
    y: np.ndarray[tuple[int], np.dtype[TNumber]],
) -> MaskedArray[tuple[int], np.dtype[TNumber]]:
    """
    Match the elements of x and y from the closest without duplicates.

    Parameters
    ----------
    x : MaskedArray[tuple[int], np.dtype[TNumber]]
        The first array of shape (n,).
    y : np.ndarray[tuple[int], np.dtype[TNumber]],
        The second array. must be smaller than x.

    Returns
    -------
    MaskedArray[tuple[int], np.dtype[TNumber]]
        (x[i], y[result[i]]) is the pair.
        If x.mask.sum() < len(y), the indices of
        the remaining y elements are added to the extra indices.

    """
    if x.ndim != 1:
        raise ValueError(f"{x.ndim=} should be 1")
    if y.ndim != 1:
        raise ValueError(f"{y.ndim=} should be 1")
    if x.shape[0] < y.shape[0]:
        raise ValueError(f"{x.shape[0]=} < {y.shape[0]=}")

    arg = np.ma.masked_array(np.full_like(x, -1, dtype=int), mask=True)
    arg[~x.mask] = argmatch_from_closest(x[~x.mask], y)
    if y.shape[0] > x.mask.sum():
        arg_new = np.asarray(list(set(np.arange(len(y))) - set(arg[~arg.mask])))
        arg[x.mask.nonzero()[0][: len(arg_new)]] = np.asarray(arg_new)
    return arg


def get_cycles(
    eigvals: Sequence[Sequence[TNumber]] | np.ndarray[tuple[int, int], np.dtype[TNumber]],
    /,
    *,
    decompose_threshold_diff_factor: float = 2,
    decompose_threshold_rate: float = 0.5,
) -> Cycles[TNumber]:
    """
    Get cycles from the eigenvalues for each point.

    Parameters
    ----------
    eigvals : Sequence[Sequence[TNumber]] | np.ndarray[tuple[int, int], np.dtype[TNumber]]
        A (ordered) sequence which contains the eigenvalues for each point.
    decompose_threshold_diff_factor : float, optional
        If the rate of elements which are closer than
        max(mean difference) * decompose_threshold_diff_factor
        is higher than decompose_threshold_rate,
        the cycle is decomposed, by default 0.5.
    decompose_threshold_rate : float, optional
        If the rate of elements which are closer than
        max(mean difference) * decompose_threshold_diff_factor
        is higher than decompose_threshold_rate,
        the cycle is decomposed, by default 0.5.

    Returns
    -------
    bool
        Whether there is a branching inside the boundary.

    """
    # cycle the eigenvalues so that
    # forall i. len(eigenvalues[0]) <= len(eigenvalues[i])
    # eigv is a list which contains the eigenvalues for each point
    eig_list: list[np.ndarray[tuple[int], np.dtype[TNumber]]] = [
        np.asarray(eigs) for eigs in eigvals
    ]
    del eigvals
    eig_max_count = np.max([len(eigs) for eigs in eig_list])

    # reorder eigenvalues so that the distances between
    # eigenvalues_sorted[i] and eigenvalues_sorted[i+1] are minimized
    # [B, eigvals]
    eigvals_c_incomp = np.ma.masked_array(
        np.full(
            (len(eig_list), eig_max_count),
            np.nan,
            dtype=np.common_type(*eig_list),
            device=eig_list[0].device,
        ),
        mask=True,
    )
    for i, eigs in enumerate(eig_list):
        if i == 0:
            eigvals_c_incomp[i, : len(eigs)] = eigs
        else:
            arg = argmatch_from_closest_masked(eigvals_c_incomp[i - 1, :], eigs)
            eigvals_c_incomp[i, ~arg.mask] = eigs[arg[~arg.mask]]

    # for each continuous eigenvalues sequences, find a sequence
    # which start is the closest to the end
    eigvals_c = eigvals_c_incomp[:, ~eigvals_c_incomp.mask.any(axis=0)]
    arg = argmatch_from_closest(eigvals_c[-1, :], eigvals_c[0, :])
    G = nx.DiGraph(list(enumerate(arg[~arg.mask])))
    cycles = list(nx.simple_cycles(G))

    # decompose cycles which are too close
    for cycle in cycles:
        mean_diff = np.max(np.mean(np.abs(np.diff(eigvals_c[:, cycle], axis=0)), axis=0))
        if np.all(
            np.mean(
                np.abs(eigvals_c[:, cycle[0]][:, None] - eigvals_c[:, cycle])
                < mean_diff * decompose_threshold_diff_factor,
                axis=0,
            )
            > decompose_threshold_rate
        ):
            G.remove_edges_from(
                [(cycle[i], cycle[(i + 1) % len(cycle)]) for i in range(len(cycle))]
            )
            G.add_edges_from([(cycle[i], cycle[i]) for i in range(len(cycle))])
    cycles = list(nx.simple_cycles(G))

    # order by length
    cycles = sorted(cycles, key=len, reverse=True)
    return Cycles(
        cycles=[eigvals_c[:, cycle].swapaxes(0, 1) for cycle in cycles],
        graph=G,
        eigvals=eigvals_c.T,
        incomplete_eigvals=eigvals_c_incomp.T,
    )



from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pytest
from numpy.typing import NDArray

from epsearch import (
    BoundaryGenerator,
    CirclesBoundary,
    RectsBoundary,
    find_branching_points_recursively,
)


@pytest.mark.parametrize(
    "boundary",
    [
        CirclesBoundary(center=0, radius=2, radius_min=0.2, n_points=128),
        RectsBoundary(center=0j, half_size=2 + 2j, half_size_min=0.2 + 0.2j, n_points_per_side=32),
    ],
)
def test_msd(
    boundary: BoundaryGenerator[Any, complex],
) -> None:
    path = Path(__file__).parent / ".cache"
    path.mkdir(exist_ok=True)
    A0 = np.asarray([[0, 0, 1, 0], [0, 0, 0, 1], [-1, 1, 0, 0], [1, -2, 0, 0]])
    A1 = np.asarray([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, -1]])

    def f(z: NDArray[Any]) -> NDArray[Any]:
        A = A0 + z[:, None, None] * A1
        return np.linalg.eigvals(A)

    def f_plot(i: int | None, k: int | None, /) -> None:
        plt.savefig(path / f"test_branching_msd_{i}_{k}_{boundary.__class__.__name__}.jpg")
        plt.clf()

    fig, _ = plt.subplots()
    res = find_branching_points_recursively(
        f,
        boundary,
        f_plot=f_plot,
        depth_first=True,
        depth_first_and_break=True,
    )
    res.plot()
    print(res.branching_points)
    print(f(np.asarray(res.branching_points)))
    fig.savefig(path / f"test_branching_msd_{boundary.__class__.__name__}.jpg")


@pytest.mark.parametrize(
    "boundary",
    [
        CirclesBoundary(center=5, radius=2, radius_min=0.2, n_points=128),
        RectsBoundary(
            center=5 + 0j, half_size=2 + 2j, half_size_min=0.2 + 0.2j, n_points_per_side=32
        ),
    ],
)
def test_random(
    boundary: BoundaryGenerator[Any, complex],
) -> None:
    N = 6
    n_params = 1
    rng = np.random.default_rng(1)
    matrices = rng.normal(size=(n_params + 1, N, N)) + 1j * rng.normal(size=(n_params + 1, N, N))

    def f(p: NDArray[Any]) -> NDArray[Any]:
        A = matrices[0] + p[:, None, None] * matrices[1]
        return np.linalg.eigvals(A)

    res = find_branching_points_recursively(f, boundary)
    res.plot()
    print(res.branching_points)
    print(f(np.asarray(res.branching_points)))
    path = Path(__file__).parent / ".cache"
    path.mkdir(exist_ok=True)
    plt.savefig(path / f"test_branching_random_{boundary.__class__.__name__}.jpg")



from functools import partial
from pathlib import Path
from typing import Any, Literal

import matplotlib.pyplot as plt
import numpy as np
import pytest
from joblib import Parallel, delayed
from numpy.typing import NDArray
from ss_hankel import SSHKwargs, ss_h_circle
from tqdm_joblib import tqdm_joblib

from epsearch._branching import BoundaryGenerator, CirclesBoundary, RectsBoundary, count_duplicate
from epsearch._diff_zeros import (
    find_branching_points_recursively_hybrid,
    find_branching_points_using_zeros_ssh,
)


@pytest.mark.parametrize(
    "hybrid,depth_first,depth_first_and_break",
    [
        (True, True, True),
        (True, True, False),
        (True, False, False),
        (False, False, False),
    ],
)
@pytest.mark.parametrize("method", ["ssh", "scipy"])
@pytest.mark.parametrize(
    "boundary",
    [
        CirclesBoundary(center=0, radius=4, radius_min=0.2, n_points=64),
        RectsBoundary(center=0j, half_size=4 + 4j, half_size_min=0.2 + 0.2j, n_points_per_side=32),
    ],
)
def test_msd(
    hybrid: bool,
    method: Literal["ssh", "scipy"],
    depth_first: bool,
    depth_first_and_break: bool,
    boundary: BoundaryGenerator[Any, complex],
) -> None:
    A0 = np.asarray([[0, 0, 1, 0], [0, 0, 0, 1], [-1, 1, 0, 0], [1, -2, 0, 0]])
    A1 = np.asarray([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, -1]])

    def f(p: NDArray[Any]) -> NDArray[Any]:
        A = A0 + p[:, None, None] * A1
        return np.linalg.eigvals(A)

    if hybrid:
        res = find_branching_points_recursively_hybrid(
            f,
            boundary,
            ssh_kwargs=SSHKwargs(circle_n_points=128),
            method=method,
            depth_first=depth_first,
            depth_first_and_break=depth_first_and_break,
        )
        print(res.branching_points)
        print(f(np.asarray(res.branching_points)))
        res.plot()
        path = Path(__file__).parent / ".cache"
        path.mkdir(exist_ok=True)
        plt.savefig(
            path / f"test_diff_msd_{hybrid}_{method}_{depth_first}_{depth_first_and_break}"
            f"_{boundary.__class__.__name__}.jpg"
        )
    else:
        if method == "scipy":
            pytest.skip("scipy not implemented yet")
        else:
            res = find_branching_points_using_zeros_ssh(  # type: ignore
                f,
                circle_n_points=128,
                circle_center=2 + 0.1j,
                circle_radius=0.6,
                max_order=5,
            )
        print(res, f(res))


@pytest.mark.skip(reason="Too heavy")
def test_random():
    path = Path(__file__).parent / ".cache"
    path.mkdir(exist_ok=True)
    N = 4
    n_order = 12
    rng = np.random.default_rng(1)
    coefs = rng.normal(size=(n_order, n_order, N, N)) + 1j * rng.normal(
        size=(n_order, n_order, N, N)
    )

    def f_plot(i: int | None, k: int | None, /) -> None:
        plt.savefig(path / f"test_diff_zeros_random_{i}_{k}.jpg")
        plt.clf()

    def f(p: NDArray[Any]) -> NDArray[Any]:
        def finner(x: NDArray[Any]) -> NDArray[Any]:
            x_ = (x.squeeze()[:, None] ** np.arange(n_order)[None, :])[:, None, :, None, None, None]
            p_ = (p[:, None] ** np.arange(n_order)[None, :])[None, :, None, :, None, None]
            A = (coefs[None, None, ...] * p_ * x_).sum(axis=2).sum(axis=2)
            return A

        eig = ss_h_circle(
            finner,
            circle_n_points=4096,
            circle_center=np.asarray([0]),
            circle_radius=np.asarray([3]),
            num_vectors=2,
            max_order=8,
        )
        return eig.eigval

    res = find_branching_points_recursively_hybrid(
        f,
        CirclesBoundary(center=0, radius=0.4, radius_min=0.001, n_points=512),
        ssh_kwargs=SSHKwargs(circle_n_points=128, max_order=4),
        f_plot=f_plot,
    )
    print(res.branching_points)
    if len(res.branching_points) > 0:
        print(f(np.asarray(res.branching_points)))
        print(count_duplicate(f(np.asarray(res.branching_points))))
    res.plot()
    plt.savefig(path / "test_diff_random.jpg")


@pytest.mark.skip(reason="Too heavy")
def test_random_linear():
    path = Path(__file__).parent / ".cache"
    path.mkdir(exist_ok=True)
    N = 4
    rng = np.random.default_rng(1)
    As = rng.normal(size=(3, N, N)) + 1j * rng.normal(size=(3, N, N))

    def f(p: NDArray[Any], q: NDArray[Any]) -> NDArray[Any]:
        p = np.asarray(p)[:, None, None]
        return np.linalg.eigvals(As[0, ...] + p * As[1, ...] + q * As[2, ...])

    def f2_plot(i: int | None, k: int | None, /) -> None:
        plt.savefig(path / f"test_diff_zeros_random_linear_{i}_{k}.jpg")
        plt.clf()

    def f2(p: NDArray[Any]) -> NDArray[Any]:
        """Return ep."""
        with tqdm_joblib(total=len(p)):
            return [
                x.branching_points
                for x in Parallel(n_jobs=-1)(
                    delayed(find_branching_points_recursively_hybrid)(
                        partial(f, q=p_),
                        CirclesBoundary(center=0, radius=10, radius_min=0.001, n_points=128),
                        ssh_kwargs=SSHKwargs(circle_n_points=128, max_order=2),
                    )
                    for p_ in p
                )
            ]

    res = find_branching_points_recursively_hybrid(
        f2,
        CirclesBoundary(center=0, radius=1, radius_min=0.001, n_points=128),
        ssh_kwargs=SSHKwargs(circle_n_points=128, max_order=2),
        f_plot=f2_plot,
    )
    res.plot()



import numpy as np
import pytest
from numpy.testing import assert_allclose
from numpy.typing import NDArray

from epsearch import get_cycles
from epsearch._cycle import argmatch_from_closest


def test_cycle():
    eigval = [
        [-5, 1, 2, 3],
        [
            3.1,
            1.1,
            2.1,
        ],
        [-1.2, 0.2, 1.2, 2.2, 3.2],
        [1.1, 2.1, 3.1],
    ]
    cycles_expected = np.asarray(
        [[1, 1.1, 1.2, 1.1], [2, 2.1, 2.2, 2.1], [3, 3.1, 3.2, 3.1]],
    )
    cycles = get_cycles(eigval)  # type: ignore
    assert cycles.eigvals.shape == (3, 4)
    assert_allclose(cycles.eigvals, cycles_expected)


def test_defective():
    eigval = [
        [1, 1],
        [1, 1, 1],
        [1, 1, 1, 1],
        [1, 1, 1],
        [1, 2],
    ]
    cycles_expected = np.asarray([[1, 1, 1, 1, 1], [1, 1, 1, 1, 2]])
    cycles = get_cycles(eigval)  # type: ignore
    assert cycles.eigvals.shape == (2, 5)
    assert_allclose(cycles.eigvals, cycles_expected)


@pytest.mark.parametrize(
    "x, y, expected",
    [
        (
            np.array([1, 2, 3]),
            np.array([2, 3, 1]),
            np.array([2, 0, 1]),
        ),
        (
            np.array([1, 2, 3]),
            np.array([1, 2, 3, 4]),
            np.array([0, 1, 2]),
        ),
    ],
)
def test_match_from_closest(
    x: NDArray[np.number], y: NDArray[np.number], expected: NDArray[np.number]
) -> None:
    result = argmatch_from_closest(x, y)
    assert_allclose(result, expected)



from collections.abc import Callable, Mapping, Sequence
from typing import Any, Generic, Literal, Protocol, TypeVar

import attrs
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from colormap_complex import colormap
from matplotlib_multicolored_line import colored_line
from numpy.typing import NDArray

from epsearch._cycle import Cycles, TNumber, get_cycles

TKey = TypeVar("TKey")
TNumber_co_ = TypeVar("TNumber_co_", bound=np.number, covariant=True)


def count_duplicate(a: NDArray[Any], /, *, eps: float = 1e-3) -> NDArray[Any]:
    """Count the number of duplicates."""
    a = np.abs(a[..., :, None] - a[..., None, :]) + np.eye(a.shape[-1]) * 1e10
    return (a.min(axis=-1) < eps).sum(axis=-1)


def contour_integral(
    values: NDArray[TNumber], /, type: Literal["circle", "square"] = "circle"
) -> NDArray[TNumber]:
    """
    Calculate contour integral.

    Parameters
    ----------
    values: NDArray[TNumber]
        Values of shape (..., num_points).
    type : Literal["circle", "square"], optional
        The type of contour, by default "circle".
        If "square", assume that first quarter of points are
        one side of the square, second quarter are the side which
        is connected to the previous side, and so on.

    Returns
    -------
    NDArray[TNumber]
        The contour integral of shape (...).

    """
    if type == "square":
        l_side = values.shape[-1] // 4
        return (
            np.mean(values[..., :l_side], axis=-1)
            + 1j * np.mean(values[..., l_side : 2 * l_side], axis=-1)
            - np.mean(values[..., 2 * l_side : 3 * l_side], axis=-1)
            - 1j * np.mean(values[..., 3 * l_side :], axis=-1)
        )
    j = np.arange(values.shape[-1]) / values.shape[-1]
    weights = np.exp(2j * np.pi * j)
    cauchy = np.mean(values * weights, axis=-1)
    return cauchy


def is_analytic(
    values: NDArray[TNumber],
    *,
    rtol: float | None = None,
    integral: NDArray[TNumber] | None = None,
    type: Literal["circle", "square"] = "circle",
) -> bool:
    """
    Check if Cauchy's integral formula is fulfilled.

    Parameters
    ----------
    values : NDArray[TNumber]
        Values of shape (..., num_points).
    rtol : float, optional
        The relative tolerance, by default 1e-3.
    integral : NDArray[TNumber] | None, optional
        The contour integral of shape (...).
    type : Literal["circle", "square"], optional
        The type of contour, by default "circle".

    Returns
    -------
    bool
        Whether Cauchy's integral formula is fulfilled
        for all sequences.

    """
    rtol = 1e-3 if rtol is None else rtol
    integral = contour_integral(values, type=type) if integral is None else integral
    return bool(np.all(np.abs(integral) < rtol * np.max(np.abs(values), axis=-1)))


class BoundaryGenerator(Protocol[TKey, TNumber_co_]):
    """A protocol for boundary generators."""

    def __call__(
        self, go_further: Mapping[TKey, bool], /
    ) -> tuple[Mapping[TKey, Sequence[TNumber_co_]], Sequence[TKey]]:
        """
        Divide-and-conquer search.

        Parameters
        ----------
        go_further : Mapping[TKey, bool]
            Whether to go further for each boundary.

        Returns
        -------
        tuple[Mapping[TKey, Sequence[TNumber_co_]], Sequence[TKey]]
            The new boundaries and the final keys.

        """
        ...

    def keys_to_values(self, keys: Sequence[TKey], /) -> Sequence[TNumber_co_]:
        """
        Get the final candidates.

        Parameters
        ----------
        keys : Sequence[TKey]
            The keys of the final candidates.

        Returns
        -------
        Sequence[TNumber_co_]
            The final candidates.

        """
        ...


@attrs.frozen(kw_only=True)
class FindExceptionalPointsRecursivelyResult(Generic[TKey, TNumber]):
    """The result of the recursive search."""

    boundaries: Mapping[TKey, Sequence[TNumber]]
    """The boundaries returned by the boundary generator."""
    eigvals: Mapping[TKey, Sequence[Sequence[TNumber]]]
    """The eigenvalues for each boundary."""
    cycles: Mapping[TKey, Cycles[TNumber]]
    """The cycles for each boundary."""
    generations: Mapping[TKey, int]
    """The generation of each boundary."""
    keys: Sequence[TKey]
    """The keys of the final candidates."""
    branching_points: Sequence[TNumber]
    """The branching points found."""
    f_boundary: BoundaryGenerator[TKey, TNumber]

    def plot(
        self,
        text_contour_integral: bool = True,
        text_additional: Callable[[Cycles[TNumber]], str] | None = None,
        set_limits: bool = True,
    ) -> None:
        """Plot the boundaries and the eigenvalues."""
        plot(
            boundaries=self.boundaries,
            cycles=self.cycles,
            generations=self.generations,
            branching_points=self.branching_points,
            text_contour_integral=text_contour_integral,
            text_additional=text_additional,
            set_limits=set_limits,
            type="circle" if isinstance(self.f_boundary, CirclesBoundary) else "square",
        )


def plot(
    *,
    boundaries: Mapping[TKey, Sequence[TNumber]] | Sequence[TNumber],
    cycles: Mapping[TKey, Cycles[TNumber]] | Cycles[TNumber],
    generations: Mapping[TKey, int] | None = None,
    branching_points: Sequence[TNumber] | None = None,
    text_contour_integral: bool = True,
    text_additional: Callable[[Cycles[TNumber]], str] | None = None,
    set_limits: bool = False,
    type: Literal["circle", "square"] = "circle",
) -> None:
    """Plot the boundaries and the eigenvalues."""
    sns.set_theme()
    _, ax = plt.subplots(2, 2, figsize=(20, 20), layout="constrained")
    ax: Sequence[plt.Axes] = ax.reshape(-1)  # type: ignore
    if isinstance(boundaries, Mapping):
        boundaries_: Mapping[Any, Sequence[TNumber]] = boundaries
    else:
        boundaries_ = {None: boundaries}
    if isinstance(cycles, Cycles):
        cycles_: Mapping[Any, Cycles[TNumber]] = {None: cycles}
    else:
        cycles_ = cycles
    del boundaries, cycles
    cmap = colormap(type="oklch")
    has_multiple_generations = (
        generations is not None and np.unique(list(generations.values())).size > 1
    )
    has_multiple = len(boundaries_) > 1
    for ik, k in enumerate(boundaries_):
        i = generations.get(k, 0) if generations is not None else 0
        boundary_ = np.asarray(boundaries_[k])
        cycle_: Cycles[TNumber] = cycles_[k]
        if not has_multiple_generations:
            color = plt.get_cmap("twilight")(np.linspace(0, 1, len(boundary_)))[:, :3]
        else:
            color = cmap(
                np.linspace(0, 1, len(boundary_)),
                1 - i / (len(boundaries_) - 1),
            )

        ax[0].scatter(
            boundary_.real,
            boundary_.imag,
            c=color,
        )
        prefix = ""
        if not has_multiple:
            prefix = f"B{ik}-{prefix}"
        if not has_multiple_generations:
            prefix = f"G{i}-{prefix}"
        if cycle_.max_cycle_length > 1:
            ax[0].text(
                boundary_[0].real,
                boundary_[0].imag,
                f"{prefix}{cycle_.max_cycle_length}",
                fontsize=8,
            )
        ax[1].scatter(
            cycle_.incomplete_eigvals.real.flatten(),
            cycle_.incomplete_eigvals.imag.flatten(),
            marker="o",
            c=np.broadcast_to(color[None, :, :], (*cycle_.incomplete_eigvals.shape[:2], 3)).reshape(
                -1, 3
            ),
        )
        for cycle in cycle_.cycles:
            colored_line(
                cycle.real.T,
                cycle.imag.T,
                c=color[:, None, :],
                ax=ax[3 if cycle.shape[0] == 1 else 2],
            )
            for j in range(cycle.shape[0]):
                text = f"{prefix}C{cycle.shape[0]}-{j}"
                if text_contour_integral and j == 0:
                    contour_integral_abs = np.abs(
                        contour_integral(
                            np.sum(cycle, axis=0) ** len(cycle)
                            - np.prod(cycle * len(cycle), axis=0),
                            type=type,
                        )
                    )
                    text += f"\n∫: {contour_integral_abs:.3g}"
                ax[3 if cycle.shape[0] == 1 else 2].text(
                    cycle[j, 0].real,
                    cycle[j, 0].imag,
                    text,
                    fontsize=8,
                )
    if branching_points is not None:
        ax[0].plot(np.real(branching_points), np.imag(branching_points), "x")
    ax[0].set_title("Trace of the parameter (p)")
    ax[0].set_xlabel("Re p")
    ax[0].set_ylabel("Im p")
    ax[1].set_title("Scatter plot of the eigenvalues (λ)")
    ax[2].set_title("Trace of the eigenvalues \nwhich period > 1 (λ)")
    ax[3].set_title("Trace of the eigenvalues \nwhich period = 1 (λ)")
    for ax_ in ax[1:]:
        ax_.set_xlabel("Re λ")
        ax_.set_ylabel("Im λ")
    if set_limits:
        xlim = ax[1].get_xlim()
        ylim = ax[1].get_ylim()
        for ax_ in ax[2:]:
            ax_.set_xlim(xlim)
            ax_.set_ylim(ylim)


def find_branching_points_recursively(
    f_eigvals: Callable[[Sequence[TNumber]], Sequence[Sequence[TNumber]]],
    f_boundary: BoundaryGenerator[TKey, TNumber],
    /,
    *,
    f_go_further: Callable[[Cycles[TNumber]], bool] | None = None,
    f_final: Callable[[Cycles[TNumber]], bool] | None = None,
    f_plot: Callable[[int | None, int | None], None] | None = None,
    eigvals_analytic: bool = True,
    rtol_analytic: float | None = None,
    depth_first: bool = False,
    depth_first_and_break: bool = False,
) -> FindExceptionalPointsRecursivelyResult[TKey, TNumber]:
    """
    Search for branching points recursively.

    Parameters
    ----------
    f_eigvals : Callable[[Sequence[TNumber]], Sequence[Sequence[TNumber]]]
        A function that takes a batch of parameters and returns the eigenvalues.
    f_boundary : BoundaryGenerator[TKey]
        A function that takes the mapping key and whether a branching point is found
        inside the boundary, and returns the new boundaries
    f_go_further : Callable[[Cycles[TNumber]], bool], optional
        A function that takes the cycles and returns whether to go further,
        by default None
    f_final : Callable[[Cycles[TNumber]], bool], optional
        A function that takes the cycles and returns whether the boundary is final,
        by default None
    f_plot : Callable[[int | None, int | None], None], optional
        A function that takes the iteration number and boundary key
        and plots the boundaries, by default None
    eigvals_analytic : bool, optional
        Whether the eigenvalues are supposed to be analytic
        on the region except for the branching points,
        by default True.
        If True, the function will check if the eigenvalues
        follow Cauchy's integral formula as well.
    rtol_analytic : float, optional
        The relative tolerance for the analytic check,
        by default None.
    depth_first : bool, optional
        Whether to use depth-first search for finding branching points,
        by default False. If True, the search will be depth-first,
        otherwise it will be breadth-first.
    depth_first_and_break : bool, optional
        Whether to use depth-first search and break
        when the first branching point is found,
        by default False. If True, the search will be depth-first
        and will break when the first branching point is found.

    Returns
    -------
    FindExceptionalPointsRecursivelyResult[TKey]
        The branching points.

    """
    if f_go_further is None:

        def f_go_further(cycles: Cycles[TNumber]) -> bool:
            return cycles.max_cycle_length > 1

    boundaries = dict(f_boundary({})[0])
    boundaries_stack: list[tuple[TKey, Sequence[TNumber]]] = list(boundaries.items())
    eigvals = {}
    cycles = {}
    generations = dict.fromkeys(boundaries.keys(), 0)
    final_keys: list[TKey] = []
    while boundaries_stack:
        # check has_branching for new boundaries
        k, boundary = boundaries_stack.pop()
        eigval = f_eigvals(boundary)
        if len(eigval) != len(boundary):
            raise ValueError(
                "f_eigvals must return the same number of eigenvalues "
                "as the number of points, but "
                f"len(f(points))={len(eigval)} != len(points)={len(boundary)}"
            )
        cycle: Cycles[TNumber] = get_cycles(eigval)
        cycles[k] = cycle
        eigvals[k] = eigval
        generation = generations[k]
        has_inside = f_go_further(cycle)
        final = has_inside and f_final(cycle) if f_final is not None else False

        print(f"Generation {generation}, key {k}: {has_inside}")
        # analytic check if eigvals_analytic is True and Circle boundary
        if isinstance(f_boundary, (CirclesBoundary, RectsBoundary)) and eigvals_analytic:
            has_inside = has_inside or not is_analytic(
                cycle.eigvals,
                rtol=rtol_analytic,
                type="circle" if isinstance(f_boundary, CirclesBoundary) else "square",
            )
        print(f"Generation {generation}, key {k}: {has_inside}")

        # plot
        if f_plot is not None:
            plot(
                boundaries=boundary,
                cycles=cycle,
                generations=None,
                type="circle" if isinstance(f_boundary, CirclesBoundary) else "square",
            )
            f_plot(generation, len(cycles))

        if final:
            final_keys.append(k)
            boundaries_new: Mapping[TKey, Sequence[TNumber]] = {}
        else:
            # get new boundaries based on the previous result
            boundaries_new, final_keys_based_on_f_boundary = f_boundary({k: has_inside})
            final_keys.extend(final_keys_based_on_f_boundary)

        # break if depth_first and depth_first_and_break
        if depth_first_and_break and final_keys:
            break

        # append new boundaries and eigenvalues
        boundaries.update(boundaries_new)
        if not depth_first:
            boundaries_stack = list(boundaries_new.items()) + boundaries_stack
        else:
            boundaries_stack.extend(boundaries_new.items())
        generations.update(dict.fromkeys(boundaries_new.keys(), generation + 1))

    result = FindExceptionalPointsRecursivelyResult(
        boundaries={k: boundary for k, boundary in boundaries.items() if k in eigvals},
        eigvals=eigvals,
        cycles=cycles,
        generations=generations,
        keys=list(dict.fromkeys(final_keys)),  # remove duplicates
        branching_points=list(f_boundary.keys_to_values(final_keys)),
        f_boundary=f_boundary,
    )
    if f_plot is not None:
        result.plot()
        f_plot(None, None)
    return result


@attrs.frozen(kw_only=True)
class Circle:
    """A circle."""

    radius: float
    """The radius of the circle."""
    center: complex
    """The center of the circle."""


@attrs.frozen(kw_only=True)
class CirclesBoundary(BoundaryGenerator[Circle, complex]):
    """
    Divide-and-conquer search using circles.

    The circles circumscribe the rectangular region.
    The search region is a square
    [Re center - radius/sqrt(2), Re center + radius/sqrt(2)]
    x [Im center - radius/sqrt(2), Im center + radius/sqrt(2)].

    Parameters
    ----------
    center : complex
        The center of the circle.
    radius : float
        The radius of the circle.
    radius_min : float
        The radius threshold to stop the recursion.
    n_points : int
        The number of points on the circle.
    extra_ratio : float, optional
        The extra ratio to enlarge the circle
        to avoid the corners of the square to
        be missed, by default 0.1.
        Must be positive or zero.

    """

    center: complex
    radius: float
    radius_min: float
    n_points: int
    extra_ratio: float = 0.1

    def _circle(self, *, center: complex, radius: float) -> tuple[Circle, Sequence[complex]]:
        points = center + radius * (1 + self.extra_ratio) * np.exp(
            2j * np.pi * np.arange(self.n_points) / self.n_points
        )
        return Circle(center=center, radius=radius), points

    def __call__(
        self, go_further: Mapping[Circle, bool], /
    ) -> tuple[Mapping[Circle, Sequence[complex]], Sequence[Circle]]:
        """
        Divide-and-conquer search using circles.

        The circles circumscribe the rectangular region.
        """
        final_keys = []
        if not go_further:
            return dict([self._circle(center=self.center, radius=self.radius)]), []
        else:
            result: dict[Circle, Sequence[complex]] = {}
            for circle, branching in go_further.items():
                if not branching:
                    continue
                if circle.radius < self.radius_min:
                    final_keys.append(circle)
                    continue
                result.update(
                    dict(
                        [
                            self._circle(
                                center=circle.center
                                + circle.radius / 2 / np.sqrt(2) * (i + j * 1j),
                                radius=circle.radius / 2,
                            )
                            for i in [-1, 1]
                            for j in [-1, 1]
                        ]
                    )
                )
            return result, final_keys

    def keys_to_values(self, keys: Sequence[Circle], /) -> Sequence[complex]:
        """
        Get the final candidates.

        Parameters
        ----------
        keys : Sequence[Circle]
            The keys of the final candidates.

        Returns
        -------
        Sequence[complex]
            The final candidates.

        """
        centers = np.asarray([circle.center for circle in keys])
        radii = np.asarray([circle.radius for circle in keys])
        result = []
        for i in range(len(centers)):
            if (np.abs(centers[i] - centers[:i]) > radii[i] + radii[:i]).all():
                result.append(centers[i])
        return result


@attrs.frozen(kw_only=True)
class Rect:
    half_size: complex
    center: complex

    @property
    def radius(self) -> float:
        return abs(self.half_size)


@attrs.frozen(kw_only=True)
class RectsBoundary(BoundaryGenerator[Rect, complex]):
    """
    Divide-and-conquer search using rectangles.

    Parameters
    ----------
    center : complex
        The center of the rectangle.
    half_size : complex
        The half size of the rectangle (width/2 + 1j * height/2).
    half_size_min : complex
        The half size threshold to stop the recursion.
    n_points_per_side : int
        The number of points per side on the rectangle.
    extra_ratio : float, optional
        The extra ratio to enlarge the rectangle
        to avoid the corners to be missed, by default 0.1.
        Must be positive or zero.

    """

    center: complex
    half_size: complex
    half_size_min: complex
    n_points_per_side: int
    extra_ratio: float = 0.1

    def _rect(self, *, center: complex, half_size: complex) -> tuple[Rect, Sequence[complex]]:
        half_size = half_size * (1 + self.extra_ratio)
        arranged = 2 * np.arange(self.n_points_per_side) / self.n_points_per_side - 1
        points = center + np.concat(
            [
                # right top
                half_size.real + 1j * half_size.imag * arranged,
                # right bottom
                1j * half_size.imag - half_size.real * arranged,
                # left bottom
                -half_size.real - 1j * half_size.imag * arranged,
                # left top
                -1j * half_size.imag + half_size.real * arranged,
            ],
            axis=0,
        )
        return Rect(center=center, half_size=half_size), points

    def __call__(
        self, go_further: Mapping[Rect, bool], /
    ) -> tuple[Mapping[Rect, Sequence[complex]], Sequence[Rect]]:
        final_keys = []
        if not go_further:
            return dict([self._rect(center=self.center, half_size=self.half_size)]), []
        else:
            result: dict[Rect, Sequence[complex]] = {}
            for rect, branching in go_further.items():
                if not branching:
                    continue
                if (
                    rect.half_size.real < self.half_size_min.real
                    and rect.half_size.imag < self.half_size_min.imag
                ):
                    final_keys.append(rect)
                    continue
                result.update(
                    dict(
                        [
                            self._rect(
                                center=rect.center
                                + rect.half_size.real / 2 * i
                                + 1j * rect.half_size.imag / 2 * j,
                                half_size=rect.half_size / 2,
                            )
                            for i in [-1, 1]
                            for j in [-1, 1]
                        ]
                    )
                )
            return result, final_keys

    def keys_to_values(self, keys: Sequence[Rect], /) -> Sequence[complex]:
        return [rect.center for rect in keys]
